import os
import re
import glob
import numpy as np
from astropy.io import fits

def create_directories(opdir, masktag, noisetag, averagetag, filtertag, savenoise, saveaverage, nofilter):
    """
    Create necessary directories for output files.

    Args:
    opdir (str): Output directory.
    masktag (str): Tag for mask files.
    noisetag (str): Tag for noise files.
    averagetag (str): Tag for averaged files.
    filtertag (str): Tag for filtered files.
    savenoise (bool): Whether to save noise files.
    saveaverage (bool): Whether to save averaged files.
    nofilter (bool): Whether to skip filtering.
    """
    if not os.path.isdir(opdir):
        os.mkdir(opdir)
    if not os.path.isdir(f'{opdir}/{masktag}'):
        os.mkdir(f'{opdir}/{masktag}')
    if savenoise and not os.path.isdir(f'{opdir}/{noisetag}'):
        os.mkdir(f'{opdir}/{noisetag}')
    if saveaverage and not os.path.isdir(f'{opdir}/{averagetag}'):
        os.mkdir(f'{opdir}/{averagetag}')
    if not nofilter and not os.path.isdir(f'{opdir}/{filtertag}'):
        os.mkdir(f'{opdir}/{filtertag}')

def get_image(fits_file):
    """
    Extract the 2D image data from a FITS file.

    Args:
    fits_file (str): Path to the FITS file.

    Returns:
    np.array: 2D image data.
    """
    input_hdu = fits.open(fits_file, ignore_missing_simple=True)[0]
    if len(input_hdu.data.shape) == 2:
        return np.array(input_hdu.data[:, :])
    elif len(input_hdu.data.shape) == 3:
        return np.array(input_hdu.data[0, :, :])
    else:
        return np.array(input_hdu.data[0, 0, :, :])

def flush_image(image_data, fits_file):
    """
    Write the 2D image data array to a FITS file.

    Args:
    image_data (np.array): 2D array of image data.
    fits_file (str): Path to the FITS file to write to.
    """
    with fits.open(fits_file, mode='update') as f:
        input_hdu = f[0]
        if len(input_hdu.data.shape) == 2:
            input_hdu.data[:, :] = image_data
        elif len(input_hdu.data.shape) == 3:
            input_hdu.data[0, :, :] = image_data
        elif len(input_hdu.data.shape) == 4:
            input_hdu.data[0, 0, :, :] = image_data
        f.flush()

def load_cube(fits_list):
    """
    Load a sequence of FITS images into a 3D numpy array.

    Args:
    fits_list (list): List of FITS file paths.

    Returns:
    np.array: 3D cube of image data.
    """
    temp = [get_image(fits_file) for fits_file in fits_list]
    return np.dstack(temp)

def natural_sort(l):
    """
    Sort the given iterable in the way that humans expect.

    Args:
    l (list): List of strings to sort.

    Returns:
    list: Naturally sorted list.
    """
    convert = lambda text: int(text) if text.isdigit() else text.lower()
    alphanum_key = lambda key: [convert(c) for c in re.split('([0-9]+)', key)]
    return sorted(l, key=alphanum_key)




import os
import sys
import glob
import time
import numpy as np
from optparse import OptionParser
from multiprocessing import Pool

from pony3d.terminal_operations import initialize_logging, hello, spacer
from pony3d.file_operations import create_directories, natural_sort
from pony3d.mask_operations import make_mask, make_averaged_mask, filter_mask, count_islands

def main():
    # Initialize logging
    logger = initialize_logging()

    # Print welcome message
    hello()
    spacer()

    # Argument parsing
    parser = OptionParser(usage='%prog [options] input_image(s)')
    # Add all the options here...
    parser.add_option('--threshold', dest='threshold', metavar='T', help='Sigma threshold for masking (default = 5.0)', default=5.0)
    parser.add_option('--boxsize', dest='boxsize', metavar='B', help='Box size to use for background noise estimation (default = 80)', default=80)
    parser.add_option('--dilate', dest='dilate', metavar='D', help='Number of iterations of binary dilation in the spatial dimensions (default = 5, set to 0 to disable)', default=3)
    parser.add_option('--specdilate', dest='specdilate', metavar='S', help='Number of iterations of binary dilation in the spectral dimension (default = 2, set to 0 to disable, filtering must be enabled)', default=2)
    parser.add_option('--chanaverage', dest='chanaverage', metavar='N', help='Width sliding channel window to use when making masks (odd numbers preferred, default = 1, i.e. no averaging)', default=1)
    parser.add_option('--saveaverage', dest='saveaverage', help='Save the result of the sliding average (default = do not save averaged image)', action='store_true', default=False)
    parser.add_option('--chanchunk', dest='chanchunk', metavar='M', help='Number of channels to load per worker when filtering single channel instances (default = 16)', default=16)
    parser.add_option('--nofilter', dest='nofilter', help='Do not filter detections for single channel instances (default = filtering enabled)', action='store_true', default=False)
    parser.add_option('--nocount', dest='nocount', help='Do not report island counts and input RMS in the log (default = report values)', action='store_true', default=False)
    parser.add_option('--savenoise', dest='savenoise', help='Enable to export noise images as FITS files (default = do not save noise images)', action='store_true', default=False)
    parser.add_option('--invert', dest='invert', help='Multiply images by -1 prior to masking (default = do not invert images)', action='store_true', default=False)
    parser.add_option('--opdir', dest='opdir', help='Name of folder for output products (default = auto generated)', default='')
    parser.add_option('--masktag', dest='masktag', help='Suffix and subfolder name for mask images (default = mask)', default='mask')
    parser.add_option('--noisetag', dest='noisetag', help='Suffix and subfolder name for noise images (default = noise)', default='noise')
    parser.add_option('--averagetag', dest='averagetag', help='Suffix and subfolder name for boxcar averaged images (default = avg)', default='avg')
    parser.add_option('--filtertag', dest='filtertag', help='Suffix and subfolder name for filtered images (default = filtered)', default='filtered')
    parser.add_option('-f', '--force', dest='overwrite', help='Overwrite existing FITS outputs (default = do not overwrite)', action='store_true')
    parser.add_option('-j', dest='j', metavar='J', help='Number of worker processes (default = 24)', default=24)

    (options, args) = parser.parse_args()

    # Parse options
    threshold = float(options.threshold)
    boxsize = int(options.boxsize)
    dilate = int(options.dilate)
    specdilate = int(options.specdilate)
    chanaverage = int(options.chanaverage)
    saveaverage = options.saveaverage
    chanchunk = int(options.chanchunk)
    nofilter = options.nofilter
    nocount = options.nocount
    savenoise = options.savenoise
    invert = options.invert
    opdir = options.opdir or f'pony3d.output.{time.strftime("%d%m%Y_%H%M%S")}'
    masktag = options.masktag
    noisetag = options.noisetag
    averagetag = options.averagetag
    filtertag = options.filtertag
    overwrite = options.overwrite
    j = int(options.j)

    pool = Pool(processes=j)

    # Create directories
    create_directories(opdir, masktag, noisetag, averagetag, filtertag, savenoise, saveaverage, nofilter)

    # Get input FITS list
    if len(args) != 1:
        logger.error('Please specify an image pattern')
        sys.exit()
    else:
        pattern = args[0]
        fits_list = natural_sort(glob.glob(f'*{pattern}*'))
        if not fits_list:
            logger.error('The specified pattern returns no files')
            sys.exit()
        else:
            nfits = len(fits_list)
            logger.info(f'Number of input images: {nfits}')

    # Report options for log
    logger.info(f'Number of worker processes: {j}')
    logger.info(f'Output folder: {opdir}')
    logger.info(f'Detection threshold: {threshold}')
    logger.info(f'Boxsize: {boxsize}')
    if dilate > 0:
        logger.info(f'Spatial dilation iteration(s): {dilate}')
    logger.info(f'Frequency averaging: {"No" if chanaverage == 1 else "Yes"}')
    if chanaverage != 1:
        logger.info(f'Average channels per worker: {chanaverage}')
        logger.info(f'Sacrificial edge channels: {chanaverage // 2}')
    logger.info(f'Single channel filtering: {"No" if nofilter else "Yes"}')
    if not nofilter:
        logger.info(f'Spectral dilation iteration(s): {specdilate}')
        logger.info(f'Filter channels per worker: {chanchunk}')
    logger.info(f'Save noise maps: {"Yes" if savenoise else "No"}')
    logger.info(f'Overwrite existing files: {"Yes" if overwrite else "No"}')

    # Make masks
    if chanaverage == 1:
        iterable_params = zip(
            fits_list, [threshold]*nfits, [boxsize]*nfits, [dilate]*nfits,
            [invert]*nfits, [opdir]*nfits, [masktag]*nfits, [noisetag]*nfits,
            [savenoise]*nfits, [overwrite]*nfits, np.arange(nfits)
        )
        pool.starmap(make_mask, iterable_params)
    else:
        input_fits_subsets = [
            fits_list[i:min(i + chanaverage, nfits)]
            for i in range(nfits - chanaverage + 1)
        ]
        logger.info(f'Sliding average will result in {len(input_fits_subsets)} output images from {nfits} inputs')
        iterable_params = zip(
            input_fits_subsets, [threshold]*len(input_fits_subsets), [boxsize]*len(input_fits_subsets), 
            [dilate]*len(input_fits_subsets), [invert]*len(input_fits_subsets), [opdir]*len(input_fits_subsets), 
            [masktag]*len(input_fits_subsets), [noisetag]*len(input_fits_subsets), [savenoise]*len(input_fits_subsets), 
            [averagetag]*len(input_fits_subsets), [saveaverage]*len(input_fits_subsets), [overwrite]*len(input_fits_subsets), 
            np.arange(len(input_fits_subsets))
        )
        pool.starmap(make_averaged_mask, iterable_params)

    # Filter masks
    if not nofilter:
        mask_list = natural_sort(glob.glob(f'{opdir}/{masktag}/*{pattern}*'))
        if not mask_list:
            logger.error('No mask images found')
            sys.exit()
        nchunks = nfits // chanchunk
        mask_subsets = [mask_list[i*chanchunk:(i+1)*chanchunk+2] for i in range(nchunks)]
        mask_subsets.append(mask_list[(nchunks-1)*chanchunk:])
        logger.info(f'Filtering masks in {len(mask_subsets)} subsets')

        iterable_params = zip(
            mask_subsets, [specdilate]*len(mask_subsets), [masktag]*len(mask_subsets), 
            [filtertag]*len(mask_subsets), [overwrite]*len(mask_subsets), np.arange(len(mask_subsets))
        )
        pool.starmap(filter_mask, iterable_params)

    # Count islands
    if not nocount:
        if not nofilter:
            mask_list = natural_sort(glob.glob(f'{opdir}/{filtertag}/*{pattern}*'))
            orig_list = [mask_fits.split('/')[-1].replace(f'.{filtertag}', '') for mask_fits in mask_list]
        else:
            mask_list = natural_sort(glob.glob(f'{opdir}/{masktag}/*{pattern}*'))
            orig_list = [mask_fits.split('/')[-1].replace(f'.{masktag}', '') for mask_fits in mask_list]

        iterable_params = zip(mask_list, orig_list, np.arange(len(mask_list)))
        pool.starmap(count_islands, iterable_params)

if __name__ == '__main__':
    main()




import numpy as np
import shutil
from scipy.ndimage import binary_dilation, minimum_filter
import scipy.special
from pony3d.file_operations import get_image, flush_image

def get_mask_and_noise(input_image, threshold, boxsize, dilate):
    """
    Generate a mask and noise image from the input image.

    Args:
    input_image (np.array): 2D input image array.
    threshold (float): Sigma threshold for masking.
    boxsize (int): Box size for background noise estimation.
    dilate (int): Number of dilation iterations.

    Returns:
    tuple: Boolean mask array and noise estimation array.
    """
    box = (boxsize, boxsize)
    n = boxsize ** 2.0
    x = np.linspace(-10, 10, 1000)
    f = 0.5 * (1.0 + scipy.special.erf(x / np.sqrt(2.0)))
    F = 1.0 - (1.0 - f) ** n
    ratio = np.abs(np.interp(0.5, F, x))
    noise_image = -minimum_filter(input_image, box) / ratio
    noise_image[noise_image < 0.0] = 1.0e-10
    median_noise = np.median(noise_image)
    noise_image[noise_image < median_noise] = median_noise
    mask_image = input_image > threshold * noise_image
    if dilate > 0:
        mask_image = binary_dilation(mask_image, iterations=dilate)
    return mask_image, noise_image

def make_mask(input_fits, threshold, boxsize, dilate, invert, opdir, masktag, noisetag, savenoise, overwrite, idx):
    """
    Create a mask for a single FITS image.

    Args:
    input_fits (str): Path to the input FITS file.
    threshold (float): Sigma threshold for masking.
    boxsize (int): Box size for background noise estimation.
    dilate (int): Number of dilation iterations.
    invert (bool): Whether to invert the image.
    opdir (str): Output directory.
    masktag (str): Tag for mask files.
    noisetag (str): Tag for noise files.
    savenoise (bool): Whether to save noise files.
    overwrite (bool): Whether to overwrite existing files.
    idx (int): Index for logging.
    """
    idx = str(idx).zfill(5)
    mask_fits = os.path.join(opdir, masktag, input_fits.replace('.fits', f'.{masktag}.fits'))
    if os.path.isfile(mask_fits) and not overwrite:
        logging.info(f'[M{idx}] Skipping {input_fits} (overwrite disabled)')
        return

    logging.info(f'[M{idx}] Reading {input_fits}')
    input_image = get_image(input_fits)
    if invert:
        logging.info(f'[M{idx}] Inverting {input_fits}')
        input_image *= -1.0
    logging.info(f'[M{idx}] Finding islands')
    mask_image, noise_image = get_mask_and_noise(input_image, threshold, boxsize, dilate)
    logging.info(f'[M{idx}] Writing {mask_fits}')
    shutil.copyfile(input_fits, mask_fits)
    flush_image(mask_image, mask_fits)
    if savenoise:
        noise_fits = os.path.join(opdir, noisetag, input_fits.replace('.fits', f'.{noisetag}.fits'))
        shutil.copyfile(input_fits, noise_fits)
        logging.info(f'[M{idx}] Writing {noise_fits}')
        flush_image(noise_image, noise_fits)

def make_averaged_mask(input_fits_subset, threshold, boxsize, dilate, invert, opdir, masktag, noisetag, savenoise, averagetag, saveaverage, overwrite, idx):
    """
    Create a mask for a sequence of averaged FITS images.

    Args:
    input_fits_subset (list): List of input FITS files.
    threshold (float): Sigma threshold for masking.
    boxsize (int): Box size for background noise estimation.
    dilate (int): Number of dilation iterations.
    invert (bool): Whether to invert the image.
    opdir (str): Output directory.
    masktag (str): Tag for mask files.
    noisetag (str): Tag for noise files.
    savenoise (bool): Whether to save noise files.
    averagetag (str): Tag for averaged files.
    saveaverage (bool): Whether to save averaged files.
    overwrite (bool): Whether to overwrite existing files.
    idx (int): Index for logging.
    """
    idx = str(idx).zfill(5)
    nfits = len(input_fits_subset)
    input_fits = input_fits_subset[nfits // 2]
    mask_fits = os.path.join(opdir, masktag, input_fits.replace('.fits', f'.{masktag}.fits'))
    if os.path.isfile(mask_fits) and not overwrite:
        logging.info(f'[A{idx}] Skipping {mask_fits} (overwrite disabled)')
        return

    logging.info(f'[A{idx}] Reading subset')
    cube = load_cube(input_fits_subset)
    mean_image = np.nanmean(cube, axis=2)
    if invert:
        logging.info(f'[M{idx}] Inverting subset')
        mean_image *= -1.0
    logging.info(f'[A{idx}] Finding islands')
    mask_image, noise_image = get_mask_and_noise(mean_image, threshold, boxsize, dilate)
    logging.info(f'[A{idx}] Writing {mask_fits}')
    shutil.copyfile(input_fits, mask_fits)
    flush_image(mask_image, mask_fits)
    if saveaverage:
        mean_fits = os.path.join(opdir, averagetag, input_fits.replace('.fits', f'.{averagetag}.fits'))
        logging.info(f'[A{idx}] Writing {mean_fits}')
        shutil.copyfile(input_fits, mean_fits)
        flush_image(mean_image, mean_fits)
    if savenoise:
        noise_fits = os.path.join(opdir, noisetag, input_fits.replace('.fits', f'.{noisetag}.fits'))
        shutil.copyfile(input_fits, noise_fits)
        logging.info(f'[A{idx}] Writing {noise_fits}')
        flush_image(noise_image, noise_fits)

import numpy as np
import shutil
from scipy.ndimage import binary_erosion, binary_dilation, binary_fill_holes
from pony3d.file_operations import get_image, flush_image, load_cube

def filter_mask(mask_subset, specdilate, masktag, filtertag, overwrite, idx):
    """
    Filter mask images for single channel islands.

    Args:
    mask_subset (list): List of mask files.
    specdilate (int): Number of iterations of binary dilation in the spectral dimension.
    masktag (str): Tag for mask files.
    filtertag (str): Tag for filtered files.
    overwrite (bool): Whether to overwrite existing files.
    idx (int): Index for logging.
    """
    idx = str(idx).zfill(5)
    template_fits = []
    output_fits = []
    exists = []

    for input_fits in mask_subset[1:-1]:
        filtered_fits = input_fits.replace(masktag, filtertag)
        template_fits.append(input_fits)
        output_fits.append(filtered_fits)
        exists.append(os.path.isfile(filtered_fits))

    if all(exists) and not overwrite:
        logging.info(f'[F{idx}] Subset is complete, skipping (overwrite disabled)')
        return

    logging.info(f'[F{idx}] Reading subset')
    cube = load_cube(mask_subset) != 0
    recon_struct = np.ones((3, 3))

    logging.info(f'[F{idx}] Filtering image')
    for i in range(cube.shape[0]):
        cut = binary_fill_holes(cube[i, :, :])
        eroded = binary_erosion(cut)
        dilated = binary_dilation(eroded, structure=recon_struct, iterations=specdilate)
        cube[i, :, :] = dilated

    for i, filtered_fits in enumerate(output_fits):
        if exists[i] and not overwrite:
            logging.info(f'[F{idx}] Skipping {filtered_fits} (overwrite disabled)')
        else:
            logging.info(f'[F{idx}] Writing {filtered_fits}')
            shutil.copyfile(template_fits[i], filtered_fits)
            flush_image(cube[:, :, i + 1], filtered_fits)

def count_islands(input_fits, orig_fits, idx):
    """
    Count the number of islands in a mask to inform cleaning.

    Args:
    input_fits (str): Path to the input mask FITS file.
    orig_fits (str): Path to the original FITS file.
    idx (int): Index for logging.
    """
    idx = str(idx).zfill(5)
    input_image = get_image(input_fits)
    input_image = input_image.byteswap().newbyteorder()  # Fix for potential endianness issue
    labeled_mask_image, n_islands = label(input_image)
    orig_image = get_image(orig_fits)
    rms = np.std(orig_image)
    logging.info(f'[C{idx}] Clean parameters: {input_fits} {n_islands} {rms}')



from setuptools import setup, find_packages

setup(
    name='pony3d',
    version='0.1.0',
    author='Ian Heywood',
    author_email='ian.heywood@physics.ox.ac.uk',
    description='Parallelised 3D mask making for spectral line deconvolution',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/IanHeywood/pony3d',
    packages=find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
    install_requires=[
        'numpy',
        'scipy',
        'astropy',
    ],
    entry_points={
        'console_scripts': [
            'pony3d=pony3d.main:main',
        ],
    },
    license='GPLv3',
)



import logging
from datetime import datetime
from pony3d import __version__

def initialize_logging():
    """
    Initialize and configure the logger.

    Returns:
    logging.Logger: Configured logger instance.
    """
    date_time = datetime.now()
    timestamp = date_time.strftime('%d%m%Y_%H%M%S')
    logfile = f'pony3d_{timestamp}.log'

    logging.basicConfig(
        filename=logfile, level=logging.DEBUG,
        format='%(asctime)s:: %(levelname)-5s :: %(message)s',
        datefmt='%d/%m/%Y %H:%M:%S ', force=True
    )
    logger = logging.getLogger()
    logger.addHandler(logging.StreamHandler())
    return logger

def hello():
    """
    Print and log the welcome message, including the current version number.
    """
    logging.info('                       .oPYo.   .oPYo.   odYo.   o    o        ')
    logging.info("                       8    8   8    8   8' `8   8    8        ")
    logging.info('                       8    8   8    8   8   8   8    8        ')
    logging.info("                       8YooP'   `YooP'   8   8   `YooP8        ")
    logging.info('                       8                              8        ')
    logging.info(f"                       8                           ooP'   v{__version__} ")

def spacer():
    """
    Add a spacer to the log for better readability.
    """
    logging.info('')
    logging.info('-' * 80)
    logging.info('')

